-- | First Order Resolution Prover
module FORP where

import Formula

import Control.Monad
import Data.List (elemIndex,group,sort,groupBy,sortBy)
import Data.List.Split (splitOn)
import Data.Maybe (mapMaybe,fromMaybe,isJust)
import Debug.Trace
import qualified Data.Set as S
import qualified Data.Map as M
import qualified Data.ByteString.Char8 as B (pack)


------------------------------------------------------------------------------
-- * order

type Order = Literal -> Literal -> Ordering

type Precedence = [PSymb]

-- | lifts a order on terms to a order on literals
mkOrder :: (Precedence -> Term -> Term -> Ordering) -> Precedence -> Order
mkOrder _ _ (NP f ts1) (PP g ts2) | f == g && isJust (variantterms ts1 ts2) = GT
mkOrder f p l1 l2
  = f p (Fun (symbol l1) (terms l1)) (Fun (symbol l2) (terms l2))

-- | constructs a lexicographic path order based on the precedence
mkLPO :: Precedence -> Order
mkLPO = mkOrder lpo

lpo :: Precedence -> Term -> Term -> Ordering
lpo _ t1 t2 | isJust $ variantterms [t1] [t2] = EQ
lpo p t1 t2 = if t1 `gtLPO` t2 then GT else LT
  where
    _ `gtLPO` (Var _) = True
    s `gtLPO`  t = s `gtLPO1` t || s `gtLPO2` t || s `gtLPO3` t

    s `gteLPO` t = s `eqT` t    || s `gtLPO`  t

    s@(Fun f _ ) `gtLPO1` (Fun g ts) = f `gtP` g && all (s `gtLPO`) ts
    _            `gtLPO1` _          = False

    s@(Fun f ss) `gtLPO2` (Fun g ts) | f == g = ss `eqTs` ts
      where
        []       `eqTs` [] = False
        (si:sis) `eqTs` (ti:tis)
          | si `eqT` ti = sis `eqTs` tis
          | otherwise   = si `gtLPO` ti && all (s `gtLPO`) tis
        _        `eqTs` _ = error "unexpected case"
    _            `gtLPO2` _            = False

    (Fun _ ss) `gtLPO3` t@(Fun _ _) = any (`gteLPO` t) ss
    _          `gtLPO3` _           = False


    f `gtP` g = error "incomplete precedence" `fromMaybe` (do
      i <- elemIndex f p
      j <- elemIndex g p
      return $ i > j)
    
    s `eqT` t = isJust $ variantterms [s] [t]

-- | given some clauses, constructs the precedence wrt to the cardinality of the symbols
mkPrecedence :: Clauses -> Precedence
mkPrecedence = reverse . fst . unzip . sortBy count . map len . group . sort . concatMap (concatMap symbols)
  where
    len zs@(z:_)      = (z,length zs)
    len _             = error "unexpected case"
    count (_,n) (_,m) = compare m n

-- | checks wether a literal is maximal wrt to some ordering and a clause
isMaximal :: Order -> Literal -> Clause -> Bool
isMaximal cmp l = not . any gt . map (`cmp` l)
  where
    gt GT = True
    gt _  = False

-- | checks wether a literal is strictly maximal wrt to some ordering and a clause
isStrictlyMaximal :: Order -> Literal -> Clause -> Bool
isStrictlyMaximal cmp l = not . any geq . map (`cmp` l)
  where
    geq LT = False
    geq _  = True


------------------------------------------------------------------------------
-- * resolution

-- | performs 'resolution' on all clauses
resolvants :: Maybe Order -> Clauses -> Clauses -> Clauses
resolvants cmp cs1 cs2 = do
  c1 <- cs1
  c2 <- cs2
  resolution cmp c1 (rename c1 c2)

-- | resolvants' cmp cs == resolvants cmp cs cs
resolvants' :: Maybe Order -> Clauses -> Clauses
resolvants' cmp cs = do
  c1 <- cs
  c2 <- cs
  guard (c1 > c2) -- ignore symmetric cases
  resolution cmp c1 (rename c1 c2)

-- | (ordered) resolution on two clauses
resolution :: Maybe Order -> Clause -> Clause -> Clauses
resolution cmpM cl1 cl2 = do
  a <- cl1
  b <- cl2
  guard $ isComplementary a b
  {-guard $ isComplementary a b && not (isNEq a) && not (isEq a)-}
  let sigma = unifyterms (terms a) (terms b)
      (c,d) = (filter (/=a) cl1, filter (/=b) cl2)
  mzero `maybe` newClause (a,c) (b,d) $ sigma
  where 
    newClause (a,c) (b,d) sigma = case cmpM of
      Nothing -> return $ substitute sigma `fmap` (c++d)
      Just cmp -> 
        let ac = (substitute sigma a, substitute sigma `fmap` c)
            bd = (substitute sigma b, substitute sigma `fmap` d) in
        if isPositive a 
            then newClause' cmp ac bd
            else newClause' cmp bd ac
    newClause' cmp (a,c) (b,d) = 
      if isStrictlyMaximal cmp a c && isMaximal cmp b d
        then return $ c ++ d else mzero

-- | performs 'factoring' on all clauses
factors :: Maybe Order -> Clauses -> Clauses
factors cmp cs = cs >>= factoring cmp

-- | (ordered) factoring on a single clause
factoring :: Maybe Order -> Clause -> Clauses
factoring cmpM cl = do
  a <- cl
  b <- cl
  guard $ isPositive a && isSimilar a b && a > b -- factoring is symmetric
  let sigma = unifyterms (terms a) (terms b)
      c = filter (\d -> d/=a && d/=b) cl
  mzero `maybe` newClause c a $ sigma 
  where 
    newClause c a sigma = case cmpM of
      Nothing  -> return $ substitute sigma `fmap` (a:c)
      Just cmp -> 
        let (a',c') = (substitute sigma a, substitute sigma `fmap` c) in
        if isStrictlyMaximal cmp a' c' then return $ a':c' else mzero 


------------------------------------------------------------------------------
-- * paramodulation

-- | checks wether a given set of clauses contains equations
isParaProblem :: Clauses -> Bool
isParaProblem cs = equality `elem` symbs || nequality `elem` symbs
  where symbs = concatMap (concatMap symbols) cs

-- | performs 'presolution' on all clauses
presolvants :: Maybe Order -> Clauses -> Clauses -> Clauses
presolvants cmp cs1 cs2 = do
  c1 <- cs1
  c2 <- cs2
  presolution cmp c1 (rename c1 c2)

-- | presolvants' cmp cs == presolvants cmp cs cs
presolvants' :: Maybe Order -> Clauses -> Clauses
presolvants' cmp cs = presolvants cmp cs cs

-- | (ordered) paramodulation of two clauses
presolution :: Maybe Order -> Clause -> Clause -> Clauses
presolution cmpM c1 c2 = do
  l1 <- c1
  l2 <- c2
  guard $ isEq l1
  let
    [s,t] = terms l1
    sigmas = mapMaybe (mkSubstitutions s t) (concatMap subterms $ terms l2)
    (c,d)  = (filter (/= l1) c1, filter (/= l2) c2)
  newClauses (l1,c) (l2,d) sigmas
  where
    mkSubstitutions s t s' = do
      sigma <- unifyterms [s] [s']
      let replace t' = if t' == s' then t else t'
      return (replace,sigma)
    newClauses (l1,c) (l2,d) = concatMap newClauses'
      where 
        newClauses' (replace,sigma) = (concatMap $ mkSubstitution sigma) (mkReplacement replace)
        mkReplacement replace = substituteeachterm' replace l2
        mkSubstitution sigma lt =
          let 
            (l1',c') = (substitute sigma l1, substitute sigma `fmap` c)
            (lt',d') = (substitute sigma lt, substitute sigma `fmap` d)
            clause = lt':c'++d'
          in case cmpM of
            Nothing -> [clause]
            Just cmp 
              | isMaximal cmp l1' d' && isMaximal cmp lt' d -> [clause]
              | otherwise                                   -> mzero

-- | performs 'pfactoring' on all clauses
pfactors :: Clauses -> Clauses
pfactors cs = cs >>= pfactoring
{-pfactors cs = filter (not . null) $ cs >>= pfactoring-}

-- | equality factoring on a single clause
pfactoring :: Clause -> Clauses
pfactoring c = do
  l1 <- c
  guard $ isNEq l1
  let [s,s'] = terms l1
      sigma = unifyterms [s] [s']
  mzero `maybe` (return . newClause l1) $ sigma
  where newClause l sigma = substitute sigma `fmap` filter (/=l) c


------------------------------------------------------------------------------
-- * simplifications

-- filter (removes elements that do not satisfy all conditions)

unaryfilter :: (Clause -> Bool) -> Clauses -> Clauses
unaryfilter = filter

-- reflexiv relations are ignored, e.g. subsumption c1 c2 returns True
binaryfilter :: (Clause -> Clause -> Bool) -> Clauses -> Clauses -> Clauses
binaryfilter f cs1 = filter allc1
  where 
    allc1 c2 = and [k c1 c2 | c1 <- cs1]
    k c1 c2 | c1 == c2  = True
            | otherwise = f c1 c2

literalOrdering :: Literal -> Literal -> Ordering
literalOrdering (PP sym1 _) (PP sym2 _) = compare sym1 sym2
literalOrdering (NP sym1 _) (NP sym2 _) = compare sym1 sym2
literalOrdering (PP _ _) (NP _ _)       = GT
literalOrdering (NP _ _) (PP _ _)       = LT
literalOrdering (PE _ _) (PE _ _)       = EQ
literalOrdering (NE _ _) (NE _ _)       = EQ
literalOrdering (PE _ _) _              = LT
literalOrdering (NE _ _) _              = LT
literalOrdering _        (PE _ _)       = LT
literalOrdering _        (NE _ _)       = LT

-- multiset inclusion wrt to ordering

data MultileqRestriction = RhsRestricted | RhsUnrestricted

-- | checks multiset inclusion wrt 'MultileqRestriction' and some ordering 
multileqBy :: Show a => MultileqRestriction -> (a -> a -> Ordering) -> [a] -> [a] -> Bool
multileqBy r f xs ys = verify (count xs) (count ys)
  where
    count = map (\zs@(z:_) -> (z, length zs)) . groupBy equals . sortBy f
    equals e1 e2 = f e1 e2 == EQ
    restricted = case r of
      RhsRestricted   -> False
      RhsUnrestricted -> True

    verify [] [] = True
    verify _  [] = False
    verify ((a,na):as) ((b,nb):bs) | a `equals` b = na <= nb && verify as bs
    verify as (_:bs) = restricted && verify as bs

multileq :: Clause -> Clause -> Bool
multileq = multileqBy RhsUnrestricted literalOrdering

-- like 'multileq' but requires psymbols c1 = psymbols c2
multileq' :: Clause -> Clause -> Bool
multileq' = multileqBy RhsRestricted literalOrdering

-- > anymatch C D
-- succeeds if a substitution s can be found st. C's c= D, where C' are all non-ground terms in C
anymatch :: Clause -> Clause -> Bool
anymatch c = or . subs c
  where
    subs c1 c2   
      | null c1 || null c2 = [True] -- no failed matching occured
    subs c1 c2 = do
      l1 <- c1
      l2 <- c2
      guard $ isSimilar l1 l2
      let 
        sigma = match l1 l2
        nc1 = filter (/= l1) c1
        nc2 = filter (/= l2) c2
      return $ maybe False (k nc1 nc2) sigma
      where k nc1 nc2 sigma = or $ subs (substitute sigma `fmap` nc1) nc2

  
-- | checks wether the given clause is a Tautology
isTautology :: Clause -> Bool
isTautology c = or $ do
  l1 <- c
  l2 <- c
  guard $ isComplementary l1 l2
  return . isJust $ variantterms (terms l1) (terms l2)

-- | given two clauses, checks wether the first subsumes the second one
-- C subsumes D, ?s. multiset of Cs is containted in multiset of D
subsumes :: Clause -> Clause -> Bool
subsumes c1 c2 = c1 `multileq` c2  && c1 `anymatch` c2

-- additionally to subsumes requires that lhs is a proper (multiple) factor of the rhs
-- TODO: not correct if we only consider positive factoring
-- | given two clauses, checks wether the first condensates the second one
condensates :: Clause -> Clause -> Bool
condensates c1 c2 = length c1 < length c2 && c1 `multileq'` c2 && c1 `anymatch` c2


------------------------------------------------------------------------------
-- * Proof

data Statement = Inconsistent | Saturated deriving Show

-- | a state consits of "old" and "new" clauses, the intuition being that "new"
-- clauses are generated from "old" clauses during a resolution step
type State = (Clauses,Clauses)

initState :: Clauses -> Step
initState cl = Continue ([], cl)

isInconsistent :: (Clauses,Clauses) -> Bool
isInconsistent (_,new) = [] `elem` new

isSaturated :: (Clauses,Clauses) -> Bool
isSaturated (_,new) = null new

data Step = 
    Continue State
  | Return   State !Statement
  deriving Show

-- | a Preprocessor is performed before any resolutionstep
type Preprocessor = Clauses -> Clauses

-- | strategies are used to alter 'State' and test for inconsistency
type Strategy = Step -> Step

data Proof = Proof !Statement State

instance Show Proof where
  show (Proof st _) = show st

-- | tries to find a proof with the default strategy (fwdSubsumption,fwdTautology)
prove :: Clauses -> Proof
{-prove = proveWith2 replacementPreprocessor fwdSubsumption-}
prove cl = proveWith2 replacementPreprocessor sty cl
  where
    sty  = if B.pack "=" `elem` prcd || B.pack "!=" `elem` prcd then sty2 else sty1
    sty1 = fwdSubsumptionSty' (Just . mkLPO $ prcd)
    sty2 = pfwdSubsumptionSty' (Just . mkLPO $ prcd)
    prcd = mkPrecedence cl

-- | tries to find a prove given a st rategy
proveWith :: Strategy -> Clauses -> Proof
proveWith = proveWith2 identityPreprocessor 

proveWith2 :: Preprocessor -> Strategy -> Clauses -> Proof
proveWith2 pre sty cl = case  mksteps . initState $ pre cl of
  Return v st' -> Proof st' v
  Continue _ -> error "assert: unexpected case"
  where
    mksteps :: Step -> Step
    mksteps (Continue (o,n)) | trace ("Continue" ++ show (length o,length n)) False = undefined
    mksteps (Return (o,n) _) | trace ("Return  " ++ show (length o, length n)) False = undefined
    mksteps r@(Return _ _) = r
    mksteps s              = sty s |> mksteps
    

------------------------------------------------------------------------------
-- * preprocessor instances

identityPreprocessor :: Preprocessor
identityPreprocessor = id

replacementPreprocessor :: Preprocessor
replacementPreprocessor cs = 
  let 
    cs2 = unaryfilter isTautology' cs 
    cs3 = binaryfilter subsumes' cs2 cs2
  in cs3
  

------------------------------------------------------------------------------
-- * strategies

-- | do nothing
identitySty :: Strategy
identitySty = id

-- | performs 'isInconsistent' and 'isSaturated' test on the state
test :: Strategy
test s = s |>> identitySty

-- | performs a (ordered) resolution step
resolutionStep :: Maybe Order -> Strategy
resolutionStep cmp = evolve
  where 
    evolve (Continue (old,new)) = Continue (old',new')
      where
        old' = old ++ new
        new' = new1 ++ new2 ++ new3

        new1 = factors cmp new
        new2 = resolvants  cmp old new
        new3 = resolvants' cmp new
    evolve r = r

-- | performs a (ordered) resolution step
presolutionStep :: Maybe Order -> Strategy
presolutionStep cmp = evolve 
  where
    evolve (Continue (old,new)) = Continue (old',new')
      where
        old' = old ++ new
        new' = new1 ++ new2 ++ new3 ++ new4 ++ new5 ++ new6

        new1 = factors cmp new
        new2 = resolvants  cmp old new
        new3 = resolvants' cmp new

        new4 = pfactors new
        new5 = presolvants  cmp old new
        new6 = presolvants' cmp new
    evolve r = r

-- normalize is not lazy
-- | remove syntactically-duplicate literals and clauses
normalize :: Strategy
normalize (Continue (old,new)) = Continue (normalize' old, normalize' new)
  where normalize' = S.toList . S.map S.toList . S.fromList . map S.fromList
normalize v = v

{-newtype AlphaLiteral = AlphaLiteral {runLit :: Literal}-}

{-instance Eq AlphaLiteral  where-}
  {-l1 == l2 = isJust $ variant' l1' (renameLiteral l1' l2')-}
    {-where (l1',l2') = (runLit l1, runLit l2)-}

{-instance Ord AlphaLiteral where-}
  {-l1 `compare` l2 -}
    {- l1 == l2  = EQ-}
    {- otherwise = runLit l1 `compare` runLit l2-}


{- like 'normalize' but considers variant of terms-}
{-normalizeModAlpha :: Strategy-}
{-normalizeModAlpha (Continue (old,new)) = Continue (normalize' old, normalize' new)-}
  {-where-}
    {-normalize' = toList . toSet-}
    {-toSet      = S.fromList . map (S.fromList . map AlphaLiteral)-}
    {-toList     = S.toList . S.map (map runLit . S.toList)-}
{-normalizeModAlpha v = v-}

  
-- lifting filters to state/strategy
isTautology' :: Clause -> Bool
isTautology' = not . isTautology

subsumes', condensates' :: Clause -> Clause -> Bool
subsumes' c    = not . subsumes c
condensates' c = not . condensates c

fwdUnary :: (Clause -> Bool) -> Strategy
fwdUnary f (Continue (old,new)) = Continue (old, unaryfilter f new)
fwdUnary _  v                   = v

rplUnary :: (Clause -> Bool) -> Strategy
rplUnary f (Continue (old,new)) = Continue (unaryfilter f old, new)
rplUnary _  v                   = v

fwdBinary :: (Clause -> Clause -> Bool)  -> Strategy
fwdBinary f (Continue (old,new)) = Continue (old, binaryfilter f old new)
fwdBinary _  v                   = v

bwdBinary :: (Clause -> Clause -> Bool)  -> Strategy
bwdBinary f (Continue (old,new)) = Continue (binaryfilter f new old, new)
bwdBinary _  v                   = v

rplBinary :: (Clause -> Clause -> Bool) -> Strategy
rplBinary f (Continue (old,new)) = Continue (binaryfilter f old old, new)
rplBinary _  v                   = v


-- | strategies wich are performed on "new" clauses wrt to "old" clauses
fwdTautology,fwdSubsumption,fwdCondensation :: Strategy
fwdTautology    = fwdUnary isTautology'
fwdSubsumption  = fwdBinary subsumes'
fwdCondensation = fwdBinary condensates'

-- | strategies which are performed on "old" clauses wrt to "new" clauses
bwdSubsumption,bwdCondensation :: Strategy
bwdSubsumption  = bwdBinary subsumes'
bwdCondensation = bwdBinary condensates'

-- | strategies which are performed on "old" clauses wrt to "old" clauses
rplTautology,rplSubsumption,rplCondensation :: Strategy
rplTautology    = rplUnary isTautology'
rplSubsumption  = rplBinary subsumes'
rplCondensation = rplBinary condensates'


------------------------------------------------------------------------------
-- * combinators

-- combinators are used combine strategies

-- pipe similar operator |> = flip ($)
-- but specialised with Continue/Return
infixl |>>,|>,|!>

-- | appends two strategies and tests for inconsistency
(|>>) :: Step -> Strategy -> Step
Continue v |>> _ | isInconsistent v = Return v Inconsistent
Continue v |>> _ | isSaturated v    = Return v Saturated
Continue v |>> f                    = f (Continue v)
Return v st |>> _                   = Return v st

-- | appends two strategies
(|>) :: Step -> Strategy -> Step
Continue v |> f                    = f (Continue v)
Return v st |> _                   = Return v st

-- | for debugging: appends two strategies and prints out size of the state
(|!>) :: Step -> Strategy -> Step
c |!> _ | trace (showc c) False = undefined
  where
    {-showc (Continue (o,n)) = "Continue" ++ show (length o,length n)-}
    showc (Continue (o,n)) = "Clauses" ++ show (length o,length n) ++ show (o,n)
    showc (Return (o,n) _) = "Return  " ++ show (length o, length n)
Continue v |!> f                    = f (Continue v)
Return v st |!> _                   = Return v st


------------------------------------------------------------------------------
-- * strategy instances
  
-- a instance should include (|>>) or test
-- normalize is not lazy

resolutionSty :: Strategy
resolutionSty = resolutionSty' Nothing

resolutionSty' :: Maybe Order -> Strategy
resolutionSty' cmp s = s |> resolutionStep cmp |>> normalize

presolutionSty :: Strategy
presolutionSty = presolutionSty' Nothing

presolutionSty' :: Maybe Order -> Strategy
presolutionSty' cmp s = s |> presolutionStep cmp |>> normalize

fwdSubsumptionSty :: Strategy
fwdSubsumptionSty = fwdSubsumptionSty' Nothing

fwdSubsumptionSty' :: Maybe Order -> Strategy
fwdSubsumptionSty' cmp s = s
  |> resolutionStep cmp |>> normalize
  |> fwdTautology|> fwdSubsumption

bwdSubsumptionSty :: Strategy
bwdSubsumptionSty = bwdSubsumptionSty' Nothing

bwdSubsumptionSty' :: Maybe Order -> Strategy
bwdSubsumptionSty' cmp s = s
  |> resolutionStep cmp |>> normalize |> bwdSubsumption

replacementSty :: Strategy
replacementSty = replacementSty' Nothing
    
replacementSty' :: Maybe Order -> Strategy
replacementSty' cmp s = s
  {-resolutionStep cmp s |>> normalize |> resolutionStep cmp |>> normalize-}
  |> resolutionStep cmp |>> normalize
  |> rplTautology |> rplSubsumption

pfwdSubsumptionSty :: Strategy
pfwdSubsumptionSty = pfwdSubsumptionSty' Nothing

pfwdSubsumptionSty' :: Maybe Order -> Strategy
pfwdSubsumptionSty' cmp s = s
  |> presolutionStep cmp |>> normalize 
  |> fwdTautology|> fwdSubsumption

--
------------------------------------------------------------------------------
-- * interpreter

strategyMap :: M.Map String (Maybe Order -> Strategy)
strategyMap = M.fromList [
    ("resolutionSty"      , resolutionSty')
  , ("fwdSubsumptionSty"  , fwdSubsumptionSty')
  , ("bwdSubsumptionSty"  , bwdSubsumptionSty')
  , ("replacementSty"     , replacementSty')

  , ("presolutionSty"     , pfwdSubsumptionSty')
  , ("pfwdSubsumptionSty" , pfwdSubsumptionSty')

  , ("identity"           , const identitySty)
  , ("test"               , const test)
  , ("normalize"          , const normalize)

  , ("resolutionStep"     , resolutionStep)
  , ("presolutionStep"    , presolutionStep)

  , ("fwdTautology"       , const fwdTautology    )
  , ("fwdSubsumption"     , const fwdSubsumption  )
  , ("fwdCondensation"    , const fwdCondensation )

  , ("bwdSubsumption"     , const bwdSubsumption  )
  , ("bwdCondensation"    , const bwdCondensation )

  , ("rplTautology"       , const rplTautology    )
  , ("rplSubsumption"     , const rplSubsumption  )
  , ("rplCondensation"    , const rplCondensation )
  ]

combinatorMap :: M.Map String (Step -> Strategy -> Step)
combinatorMap = M.fromList [
    ("|>", (|>))
  , ("|>>", (|>>))
  , ("||>", (|!>))
  ]

preprocessorMap :: M.Map String Preprocessor
preprocessorMap = M.fromList [
    ("identityPreprocessor", identityPreprocessor)
  , ("replacementPreprocessor", replacementPreprocessor)
  ]

isCmd :: String -> Bool
isCmd cmd = case splitOn "*" cmd of
  [_,_,_] -> True
  _ -> False


interpret :: String -> (Preprocessor, Clauses -> Strategy)
interpret cmd = case splitOn "*" cmd of
  [cmp, pp, sty] -> (mkPP pp, mkSty (mkOrd cmp) sty)
  _ -> error "invalid: command"
  where
    mkOrd ""     = const Nothing
    mkOrd "none" = const Nothing
    mkOrd "some" = Just . mkLPO . mkPrecedence
    mkOrd fs     = const $ Just $ mkLPO (map B.pack $ splitOn ">" fs) 

    mkPP "" = replacementPreprocessor
    mkPP pp = preprocessorMap M.! pp

    mkSty ord ""  cl = fwdSubsumptionSty' (ord cl)
    mkSty ord sty cl = composeSty . reverse . words $ sty 
      where
        cmp = ord cl
        composeSty []       = error "FORP.compose: empty list"
        composeSty [s]      = strategyMap M.! s $ cmp
        composeSty (s:c:xs) = composeSty xs . flip com sty
          where
            sty = strategyMap M.! s $ cmp
            com = combinatorMap M.! c


