{-# LANGUAGE OverloadedStrings, TypeSynonymInstances, FlexibleInstances, OverlappingInstances #-}

-- | Terms, Literal, Clauses
module Formula 
where

import Data.List (foldl')
import Text.PrettyPrint.ANSI.Leijen

import Text.Parsec.String (Parser)
import Text.Parsec.Language (emptyDef)
import Text.Parsec as P hiding (char,string)
import qualified Text.Parsec.Token as T
{-import Debug.Trace-}

{-import Data.ByteString (ByteString)-}
import qualified Data.ByteString.Char8 as B
import qualified Data.Set as S
import Data.Maybe (fromMaybe)
import Control.Monad (mplus)



-----------------------------------------------------------------------------
-- * utility functions --------------------------------------------------------

prettyxs :: Pretty a => [a] -> Doc
prettyxs as | null as = empty
prettyxs as           = encloseSep lparen rparen comma (map pretty as)


-----------------------------------------------------------------------------
-- * terms 

type FSymb = B.ByteString
type VSymb = B.ByteString

data Term =
    Var !VSymb
  | Fun !FSymb [Term]
  deriving (Eq, Ord)

instance Pretty Term where
  pretty (Var v)    = text (B.unpack v)
  pretty (Fun a ts) = text (B.unpack a) <> prettyxs ts

instance Show Term where
  show = show . pretty


var :: VSymb -> Term
var = Var

fun :: FSymb -> [Term] -> Term 
fun = Fun

constant :: FSymb -> Term
constant s = fun s []

mapterm :: (Term -> Term) -> Term -> Term
mapterm f (Fun a ts) = f (Fun a (map (mapterm f) ts))
mapterm f v          = f v

-- | apply function to each subterm individually
mapeachterm :: (Term -> Term) -> Term -> [Term]
mapeachterm f (Fun a ts) = f (Fun a ts) : map (Fun a) (k [] ts)
  where 
    k [] []       = []
    k ts1 (t:ts2) = map (\lt -> ts1 ++ lt :ts2) (mapeachterm f t) ++ k (ts1 ++ [t]) ts2
    k _  []       = []
mapeachterm f v = [f v]

mapvars :: (Term -> Term) -> Term -> Term
mapvars f (Fun a ts)  = Fun a (map (mapvars f) ts)
mapvars f v           = f v

foldterm :: (Term -> a -> a) -> a -> Term -> a
foldterm f z (Fun a ts)  = 
  let z' = foldl' (foldterm f) z ts in f (Fun a ts) z'
foldterm f z v = f v z

-- | return all subterms of a term
subterms :: Term -> [Term]
subterms = foldterm (:) []

isVar :: Term -> Bool
isVar (Var _) = True
isVar _       = False

-- | return all variables of a term
termvars :: Term -> [Term]
termvars = foldterm (\t -> if isVar t then (t:) else id) []

isFun :: Term -> Bool
isFun (Fun _ _) = True
isFun _         = False

-- | return all functions of a term
termfuns :: Term -> [Term]
termfuns = foldterm (\t -> if isFun t then (t:) else id) []

termsymb :: Term -> FSymb
termsymb (Fun s _) = s
termsymb (Var s)   = s


-----------------------------------------------------------------------------
-- ** substitution

-- | a substitution as function from terms to terms
type Substitution = Term -> Term

-- | applies a substituteion to a term
substituteterms :: Substitution -> Term -> Term
substituteterms = mapterm

-- | returns a list of terms where the given substitution is applied on each subterm
substituteeachterm :: Substitution -> Term -> [Term]
substituteeachterm = mapeachterm

substitutevars :: Substitution -> Term -> Term
substitutevars = mapvars

emptysubstitution :: Substitution
emptysubstitution = id

compose :: Substitution -> Substitution -> Substitution
compose sigma tau = tau . sigma

data MatchKind = Match | Unify | Variant

-- | returns the substitution if the test for matching, unification, variance is successfull,
-- otherwise Nothing
matchterms, unifyterms,variantterms :: [Term] -> [Term] -> Maybe Substitution
(matchterms,unifyterms,variantterms) = (
    unifyts Match (Just emptysubstitution)
  , unifyts Unify (Just emptysubstitution)
  , unifyts Variant (Just emptysubstitution)
  )
  where 
    unifyt Variant v1@(Var _) v2@(Var _) = Just (\v -> if v == v1 then v2 else v)
    unifyt Variant (Var _) _       = Nothing
    unifyt Variant _       (Var _) = Nothing


    unifyt _ (Var x) (Var y) 
      | x == y = Just id
    unifyt _ v1@(Var _) f    
      | v1 `elem` termvars f = Nothing
    unifyt _ v1@(Var _) f = Just (\t -> if t == v1 then f else t)

    unifyt Unify f (Var y) = unifyt Unify (Var y) f
    unifyt Match _ (Var _) = Nothing

    unifyt b (Fun f1 ts1) (Fun f2 ts2) 
      | f1 == f2 = unifyts b (Just emptysubstitution) ts1 ts2
      | otherwise = Nothing
        
    unifyts _ s0M [] [] = s0M
    unifyts b s0M (t1:ts1) (t2:ts2) = do
      let subst = substitutevars
      s0 <- s0M 
      s1 <- unifyt b (subst s0 t1) (subst s0 t2)
      let s2 = return $ compose s0 s1
      unifyts b s2 ts1 ts2
    unifyts _ _ ts1 ts2 = error $ "unifyterms: unexpected case: " ++ show (ts1,ts2)


------------------------------------------------------------------------------
-- * clauses

data Literal =
    PP !FSymb [Term]
  | NP !FSymb [Term]
  | PE Term Term
  | NE Term Term
  deriving (Eq, Ord)

instance Pretty Literal where
  pretty (PP symb ts) = string (B.unpack symb) <> encloseSep lparen rparen comma (map pretty ts)
  pretty (NP symb ts) = char '~' <> string (B.unpack symb) <> encloseSep lparen rparen comma (map pretty ts)
  pretty (PE t1 t2) = pretty t1 <> char '=' <> pretty t2
  pretty (NE t1 t2) = pretty t1 <> text "!=" <> pretty t2

instance Show Literal where
  show = show . pretty

type Clause = [Literal]

emptyClause :: Clause
emptyClause = []

type Clauses = [Clause]

type PSymb = B.ByteString

equality :: B.ByteString
equality = B.pack "="

nequality :: B.ByteString
nequality = B.pack "!="

symbol :: Literal -> FSymb
symbol (PP symb _) = symb
symbol (NP symb _) = symb
symbol (PE _  _) = "="
symbol (NE _  _) = "!="

-- | return the predicate and function symbols of a literal
symbols :: Literal -> [FSymb]
symbols l = symbol l : (map termsymb . concatMap termfuns $ terms l)

terms :: Literal -> [Term]
terms (PP _ ts) = ts
terms (NP _ ts) = ts
terms (PE s t) = [s,t]
terms (NE s t) = [s,t]

-- | checks if the orientation and the symbol of two literals is the same
isSimilar :: Literal -> Literal -> Bool
isSimilar (PP symb1 _) (PP symb2 _) | symb1 == symb2 = True
isSimilar (NP symb1 _) (NP symb2 _) | symb1 == symb2 = True
isSimilar (PE symb1 _) (PE symb2 _) | symb1 == symb2 = True
isSimilar (NE symb1 _) (NE symb2 _) | symb1 == symb2 = True
isSimilar _ _                                        = False

-- | checks if the orientation of two literals is different and its symbols are equal
isComplementary :: Literal -> Literal -> Bool
isComplementary (PP symb1 _) (NP symb2 _) | symb1 == symb2 = True
isComplementary (NP symb1 _) (PP symb2 _) | symb1 == symb2 = True
isComplementary (PE _  _)    (NE _  _)                    = True
isComplementary (NE _  _)    (PE _  _)                    = True
isComplementary _ _                                       = False

isPositive :: Literal -> Bool
isPositive (PP _ _) = True
isPositive (NP _ _) = False
isPositive (PE _ _) = True
isPositive (NE _ _) = False

isNegative :: Literal -> Bool
isNegative = not . isPositive

isEq :: Literal -> Bool
isEq (PE _ _) = True
isEq _        = False

isNEq :: Literal -> Bool
isNEq (NE _ _) = True
isNEq _        = False

mapliteral :: (Literal -> Literal) -> (Term -> Term) -> Literal -> Literal
mapliteral f g (PP sym t) = f (PP sym $ map g t)
mapliteral f g (NP sym t) = f (NP sym $ map g t)
mapliteral f g (PE t1 t2) = f (PE (g t1) (g t2))
mapliteral f g (NE t1 t2) = f (NE (g t1) (g t2))

-- | lift 'substituteterm' to literals
substitute :: Substitution -> Literal -> Literal
substitute sigma = mapliteral id (substitutevars sigma)
--
-- | lifts 'substituteeachterm' to literals
substituteeachterm' :: Substitution -> Literal -> [Literal]
substituteeachterm' = mapeachterm'

-- | lift 'mapeachterm' to literals
mapeachterm' :: (Term -> Term) -> Literal -> [Literal]
mapeachterm' f l = case l of
  PP a ts -> map (PP a) (k [] ts)
  NP a ts -> map (NP a) (k [] ts)
  PE t1 t2 -> [ PE t1' t2 | t1' <- mapeachterm f t1] ++ [ PE t1 t2' | t2' <- mapeachterm f t2]
  NE t1 t2 -> [ NE t1' t2 | t1' <- mapeachterm f t1] ++ [ NE t1 t2' | t2' <- mapeachterm f t2]
  where 
    k [] []       = []
    k ts1 (t:ts2) = map (\lt -> ts1 ++ lt :ts2) (mapeachterm f t) ++ k (ts1 ++ [t]) ts2
    k _  []       = []

-- | lift 'unifyterms' to literals
-- does not differentiate between "=" and "!="
-- does not check orientation
unify :: Literal -> Literal -> Maybe Substitution
unify = unify' unifyterms
  where 
    unify' f l1 l2  
      | f1 == f2  = f ts1 ts2
      | otherwise = Nothing
      where 
        (f1,ts1) = (symbol' l1, terms l1)
        (f2,ts2) = (symbol' l2, terms l2)
    symbol' (PE _ _) = "="
    symbol' (NE _ _) = "="
    symbol' l        = symbol l

-- | lift 'matchterms' to literals
-- does not check orientation
match :: Literal -> Literal -> Maybe Substitution
match l1 l2 =
  if symbol l1 == symbol l2 then matchterms (terms l1) (terms l2) else Nothing
  
-- | same as 'match' but checks if the orientation and the symbols of the
-- given literals are equal
match' :: Literal -> Literal -> Maybe Substitution
match' l1 l2 
  | not $ isSimilar l1 l2 = Nothing
  | otherwise             = if symbol l1 == symbol l2 then matchterms (terms l1) (terms l2) else Nothing

-- | lift 'variantterms' to literals
-- does not check orientation
variant :: Literal -> Literal -> Maybe Substitution
variant l1 l2
  | symbol l1 /= symbol l2 = Nothing
  | isEq l1 || isNEq l1   = variantterms ts1 ts2 `mplus` variantterms ts2 ts1
  | otherwise             = variantterms ts1 ts2
  where (ts1,ts2) = (terms l1, terms l2)

-- | same as 'variant' but checks if the orientation and the symbols of the
-- given literals are equal
variant' :: Literal -> Literal -> Maybe Substitution
variant' l1 l2 
  | not $ isSimilar l1 l2 = Nothing
  | isEq l1 || isNEq l1   = variantterms ts1 ts2 `mplus` variantterms ts2 ts1
  | otherwise             = variantterms ts1 ts2
  where (ts1,ts2) = (terms l1, terms l2)

-- | given two literals, renames the second one such that the variables differ from the first one
renameLiteral :: Literal -> Literal -> Literal
renameLiteral l1 l2 
  | S.null common = l2
  | otherwise     = substitute (lookupDefault varmap ) l2
  where
    common = vars1 `S.intersection` vars2
    vars1 = vars l1
    vars2 = vars l2
    vars = S.fromList . concatMap termvars . terms
    fresh v = if v `S.member` common then fresh' $ mkVar '#' v else v
    fresh' v = if v `S.member` common || v `S.member` vars2 then fresh' $ mkVar '$' v else v
    varmap = map (\v -> (v,fresh v)) $ S.toList vars2
    lookupDefault m v = v `fromMaybe` lookup v m
    mkVar c (Var x) = Var $ B.cons c x
    mkVar _ t       = t


-- | given two clauses, renames the second clause such that the variables differ from the first one
rename :: Clause -> Clause -> Clause
rename c1 c2
  | S.null common = c2
  | otherwise     = substitute (lookupDefault varmap ) `fmap` c2
  where
    common = vars1 `S.intersection` vars2
    vars1 = vars c1
    vars2 = vars c2
    vars = S.fromList . concatMap termvars . concatMap terms
    fresh v = if v `S.member` common then fresh' $ mkVar '#' v else v
    fresh' v = if v `S.member` common || v `S.member` vars2 then fresh' $ mkVar '$' v else v
    varmap = map (\v -> (v,fresh v)) $ S.toList vars2
    lookupDefault m v = v `fromMaybe` lookup v m
    mkVar c (Var x) = Var $ B.cons c x
    mkVar _ t       = t


------------------------------------------------------------------------------
-- * Parser 
-- parses cnf format of TPDP

parseProblem :: String -> Clauses
parseProblem = parseFromTPTP

parseAny :: Parser a -> String -> a 
parseAny parser s = case parse parser "" s of
  Left err -> error $ "parse  error: " ++ show err
  Right cl -> cl

-- a restricted parser for the TPTP syntax
parseFromTPTP :: String -> Clauses
parseFromTPTP = parseAny parseTPTP

lexer :: T.TokenParser ()
lexer = T.makeTokenParser emptyDef {
    T.identStart    = letter
  , T.identLetter   = alphaNum <|> oneOf "_"
  , T.reservedNames = [
        "cnf" 
      --, "conjecture"
      --, "hypothesis"
      --, "axiom"
      --, "negated_conjecture"
      ]
  , T.caseSensitive = True
  , T.commentLine   = "%"
}


whiteSpace :: Parser ()
whiteSpace = T.whiteSpace lexer
identifier :: Parser String
identifier = T.identifier lexer
reserved   :: String -> Parser ()
reserved   = T.reserved lexer
tsymbol    :: String -> Parser String
tsymbol    = T.symbol lexer
tcolon     :: Parser String
tcolon     = T.colon lexer
tcomma     :: Parser String
tcomma     = T.comma lexer
tdot       :: Parser String
tdot       = T.dot lexer
charLiteral :: Parser Char
charLiteral = T.charLiteral lexer

tparens     = T.parens lexer


parseTPTP :: Parser Clauses
parseTPTP = many parseCNF

parseCNF :: Parser Clause
parseCNF = do
  whiteSpace
  reserved "cnf"
  cl <- tparens body
  tdot
  return cl
  <?> "parseCNF"
  where 
    body = do
      identifier >> tcomma
      identifier >> tcomma
      parseFormula

parseFormula :: Parser Clause
parseFormula = tparens parseDisjunction <|> parseDisjunction <?> "parseFormula"

parseDisjunction :: Parser Clause
parseDisjunction = parseLiteral `sepBy` tsymbol "|"

parseLiteral :: Parser Literal
parseLiteral = do
  tsymbol "~"
  sym <- parseSymb
  ts  <- parseTerms
  return $ NP (B.pack sym) ts
  <|> try (do
  t1 <- parseTerm
  whiteSpace
  op <- tsymbol "=" <|> tsymbol "!="
  whiteSpace
  t2 <- parseTerm
  return $ if op == "=" then PE t1 t2 else NE t1 t2
  )
  <|> do
  sym <- parseSymb
  ts  <- parseTerms
  return $ PP (B.pack sym) ts
  <?> "parseLiteral"

parseSymb :: Parser String
parseSymb = do
  c  <- lower
  cs <- option [] identifier
  return $ c:cs

parseTerms :: Parser [Term]
parseTerms = do
  try (tsymbol "(")
  ts <- parseTerm `sepBy` tcomma
  tsymbol ")"
  return ts
  <|>
  return []
  <?> "parseTerms"

parseTerm :: Parser Term
parseTerm = parseVariable <|> parseFunction <?> "parseTerm"

parseVariable :: Parser Term
parseVariable = do
  c  <- upper
  cs <- option [] identifier
  return $ Var (B.pack $ c:cs)

parseFunction :: Parser Term
parseFunction = do
  sym <- parseSymb
  ts  <- parseTerms
  return $ Fun (B.pack sym) ts


